@@ -1,6 +1,6 @@
 use std::io::{self, Write};
 use std::path::PathBuf;
-use clap::{Command, Arg};
+use clap::{Command, Arg, crate_version};
 
 use sphinx::frontend;
 use sphinx::source::{ModuleSource, SourceText};
@@ -19,7 +19,7 @@ fn main() {
     env_logger::init();
     
     let app = Command::new("sphinx")
-        .version("0.0")
+        .version(crate_version!())
         .author("M. Werezak <mwerezak@gmail.com>")
         .about("An interpreter for the Sphinx programming language")
         .arg(
@@ -286,7 +286,7 @@ impl Repl {
         // bind the result expression to a global name
         let result_name = interner.get_or_intern("_");
         let result_decl = Expr::Assignment(Box::new(Assignment {
-            assign: AssignType::DeclImmutable,
+            modifier: AssignType::DeclImmutable,
             lhs: LValue::Identifier(result_name),
             rhs: result_expr,
             op: None,
