@@ -74,6 +74,11 @@ pub fn create_primitive_ctors(env: Gc<NamespaceEnv>) {
         Ok(Variant::from(value.as_float()?))
     });
     
+    // convert a value into a string
+    let as_str = native_function!(str, env, params(value) => {
+        Ok(Variant::from(value.fmt_str()?))
+    });
+    
     // marker type constructor
     // let marker = native_function!(marker, env, params(marker) => {
     //     let symbol = marker.as_strval()
@@ -88,6 +93,7 @@ pub fn create_primitive_ctors(env: Gc<NamespaceEnv>) {
         fun _ = as_bits;
         fun _ = as_int;
         fun _ = as_float;
+        fun _ = as_str;
     });
 }
 
@@ -105,7 +111,7 @@ pub fn create_metamethod_builtins(env: Gc<NamespaceEnv>) {
     
     // produces a tuple (item, next_state) for a given iterator state
     let next = native_function!(next, env, params(value), defaults(state = Variant::Nil) => {
-        let result = vec![
+        let result = vec![ 
             value.iter_get(state)?,
             value.iter_next(state)?,
         ];
@@ -118,8 +124,8 @@ pub fn create_metamethod_builtins(env: Gc<NamespaceEnv>) {
         let iter = value.iter_init()?;
         
         let result = vec![
-            iter.iter,
-            iter.state,
+            *iter.get_iter(),
+            *iter.get_state(),
         ];
         
         Ok(Variant::from(result.into_boxed_slice()))
