@@ -432,10 +432,9 @@ impl From<&Scope> for ScopeDrop {
 }
 
 impl CodeGenerator<'_> {
-    fn emit_begin_scope(&mut self, symbol: Option<&DebugSymbol>, tag: ScopeTag, label: Option<&Label>) -> &mut Scope {
+    fn emit_begin_scope(&mut self, symbol: Option<&DebugSymbol>, label: Option<&Label>, tag: ScopeTag) -> &mut Scope {
         let chunk_id = self.chunk_id;
-        self.scopes_mut().push_scope(symbol, tag, label.copied());
-        self.scopes_mut().local_scope_mut().unwrap()
+        self.scopes_mut().push_scope(symbol, label.copied(), tag)
     }
     
     fn emit_end_scope(&mut self) -> Scope {
@@ -479,18 +478,21 @@ impl CodeGenerator<'_> {
     // If the local name cannot be found, no instructions are emitted and None is returned
     fn try_emit_load_local(&mut self, symbol: Option<&DebugSymbol>, name: &LocalName) -> Option<u16> {
         if let Some(index) = self.scopes().resolve_local(name).map(|local| local.index()) {
-            if let Ok(index) = u8::try_from(index) {
-                self.emit_instr_byte(symbol, OpCode::LoadLocal, index);
-            } else {
-                self.emit_instr_data(symbol, OpCode::LoadLocal16, &index.to_le_bytes());
-            }
-            
+            self.emit_load_local_index(symbol, index);
             Some(index)
         } else{
             None
         }
     }
     
+    fn emit_load_local_index(&mut self, symbol: Option<&DebugSymbol>, index: LocalIndex) {
+        if let Ok(index) = u8::try_from(index) {
+            self.emit_instr_byte(symbol, OpCode::LoadLocal, index);
+        } else {
+            self.emit_instr_data(symbol, OpCode::LoadLocal16, &index.to_le_bytes());
+        }
+    }
+    
     fn try_emit_load_upval(&mut self, symbol: Option<&DebugSymbol>, name: &LocalName) -> CompileResult<Option<u16>> {
         if let Some(index) = self.scopes_mut().resolve_or_create_upval(name)?.map(|upval| upval.index()) {
             if let Ok(index) = u8::try_from(index) {
@@ -512,6 +514,22 @@ impl CodeGenerator<'_> {
             self.emit_instr_data(symbol, OpCode::CloseUpvalue16, &index.to_le_bytes());
         }
     }
+    
+    fn emit_create_temporary(&mut self, symbol: Option<&DebugSymbol>, access: Access) -> CompileResult<LocalIndex> {
+        debug_assert!(self.scopes().is_temporary_scope());
+        
+        match self.scopes_mut().insert_local(access, LocalName::Anonymous)? {
+            InsertLocal::CreateNew(local_index) => {
+                self.emit_instr(symbol, OpCode::InsertLocal);
+                Ok(local_index)
+            }
+            
+            InsertLocal::HideExisting(local_index) => {
+                self.emit_assign_local(symbol, local_index);
+                Ok(local_index)
+            }
+        }
+    }
 }
 
 
@@ -713,7 +731,7 @@ impl CodeGenerator<'_> {
         
         let loop_target = self.current_offset();
         
-        self.emit_begin_scope(symbol, ScopeTag::Loop, label);
+        self.emit_begin_scope(symbol, label, ScopeTag::Loop);
         
         self.compile_stmt_block(body)?;
         let loop_scope = self.emit_end_scope();
@@ -738,7 +756,7 @@ impl CodeGenerator<'_> {
         
         let loop_target = self.current_offset();
         
-        self.emit_begin_scope(symbol, ScopeTag::Loop, label);
+        self.emit_begin_scope(symbol, label, ScopeTag::Loop);
         self.compile_stmt_block(body)?;
         let loop_scope = self.emit_end_scope();
         
@@ -758,7 +776,7 @@ impl CodeGenerator<'_> {
     
     fn compile_for_loop(&mut self, symbol: Option<&DebugSymbol>, label: Option<&Label>, lvalue: &LValue, iter: &Expr, body: &StmtList) -> CompileResult<()> {
         
-        self.emit_begin_scope(symbol, ScopeTag::Loop, label);
+        self.emit_begin_scope(symbol, label, ScopeTag::Loop);
         
         // initialize iterator
         self.compile_expr(symbol, iter)?;
@@ -797,6 +815,14 @@ impl CodeGenerator<'_> {
 }
 
 ///////// Expressions /////////
+
+// helper to indicate if the length of a sequence is statically known or if it is stored in a local
+enum Unpack {
+    Empty,
+    Static(IntType),
+    Dynamic,
+}
+
 impl CodeGenerator<'_> {
     fn compile_expr_with_symbol(&mut self, expr: &ExprMeta) -> CompileResult<()> {
         let symbol = expr.debug_symbol();
@@ -819,42 +845,144 @@ impl CodeGenerator<'_> {
             
             Expr::Assignment(assign) => {
                 if let Some(op) = assign.op {
-                    self.compile_update_assignment(symbol, op, assign.assign, &assign.lhs, &assign.rhs)?;
+                    self.compile_update_assignment(symbol, op, assign.modifier, &assign.lhs, &assign.rhs)?;
                 } else {
                     self.compile_expr(symbol, &assign.rhs)?;
-                    self.compile_assignment(symbol, assign.assign, &assign.lhs)?;
+                    self.compile_assignment(symbol, assign.modifier, &assign.lhs)?;
                 }
             },
             
             Expr::Tuple(items) => self.compile_tuple(symbol, items)?,
             
-            Expr::Ellipsis(..) => return Err("\"...\" is not allowed here".into()),
+            // unpacking is only allowed in invocation, tuple literals, and by itself in parentheses
+            // note: assignment uses *packing*, not unpacking, which is the LValue dual of packing.
+            Expr::Unpack(Some(..)) => return Err("unpack expression must be enclosed in parentheses".into()),
+            Expr::Unpack(None) => return Err("\"...\" is not allowed here".into()),
             
             Expr::Block { label, suite } => self.compile_block_expression(symbol, label.as_ref(), suite)?,
             Expr::IfExpr { branches, else_clause } => self.compile_if_expression(symbol, branches, else_clause.as_ref().map(|expr| &**expr))?,
             
             Expr::FunctionDef(fundef) => self.compile_function_def(symbol, fundef)?,
-            
-            Expr::Echo(expr) => {
-                self.compile_expr(symbol, expr)?;
-                self.emit_instr(symbol, OpCode::Inspect);
-            },
         }
         Ok(())
     }
     
     fn compile_tuple(&mut self, symbol: Option<&DebugSymbol>, expr_list: &[ExprMeta]) -> CompileResult<()> {
-        let len = u8::try_from(expr_list.len())
-            .map_err(|_| "tuple length limit exceeded")?;
-        
-        for expr in expr_list.iter() {
-            self.compile_expr_with_symbol(expr)?;
+        match self.compile_unpack_sequence(symbol, expr_list)? {
+            Unpack::Empty => self.emit_instr(symbol, OpCode::Empty),
+            
+            Unpack::Static(len) => {
+                if let Ok(len) = u8::try_from(len) {
+                    self.emit_instr_byte(symbol, OpCode::Tuple, len);
+                } else {
+                    self.compile_integer(symbol, len)?;
+                    self.emit_instr(symbol, OpCode::TupleN);
+                }
+            }
+            
+            Unpack::Dynamic => {
+                self.emit_instr(symbol, OpCode::TupleN);
+            }
         }
-        
-        self.emit_instr_byte(symbol, OpCode::Tuple, len);
         Ok(())
     }
     
+    // compiles to a sequence of values
+    fn compile_unpack_sequence(&mut self, symbol: Option<&DebugSymbol>, seq: &[ExprMeta]) -> CompileResult<Unpack> {
+        if seq.is_empty() {
+            return Ok(Unpack::Empty)
+        }
+        
+        let (last, rest) = seq.split_last().unwrap();
+        
+        let mut static_len: IntType = 0;
+        let mut unpack_len = None; // accumulate length in an anonymous temporary if needed
+        
+        for expr in rest.iter() {
+            match expr.variant() {
+                Expr::Unpack(None) => return Err("need a value to unpack".into()),
+                
+                Expr::Unpack(Some(unpack)) => {
+                    let symbol = Some(expr.debug_symbol());
+                    
+                    self.compile_expr(symbol, unpack)?;
+                    self.emit_instr(symbol, OpCode::IterInit);
+                    self.emit_instr(symbol, OpCode::IterUnpack);
+                    
+                    // store unpack len in accumulator
+                    if let Some(local_index) = unpack_len {
+                        self.emit_load_local_index(symbol, local_index);
+                        self.emit_instr(symbol, OpCode::Add);
+                        self.emit_assign_local(symbol, local_index);
+                    } else {
+                        self.emit_begin_scope(symbol, None, ScopeTag::Temporary);
+                        let local_index = self.emit_create_temporary(symbol, Access::ReadWrite)?;
+                        unpack_len = Some(local_index);
+                    }
+                    
+                    self.emit_instr(symbol, OpCode::Pop);
+                }
+                
+                _ => {
+                    self.compile_expr_with_symbol(expr)?;
+                    static_len = static_len.checked_add(1)
+                        .ok_or("unpack length limit exceeded")?;
+                }
+            }
+        }
+        
+        // if the last item is an unpack expression, it does not need to use the local accumulator
+        // TODO should there be a dedicated accumulator register?
+        match last.variant() {
+            Expr::Unpack(None) => return Err("need a value to unpack".into()),
+            
+            Expr::Unpack(Some(unpack)) => {
+                let symbol = Some(last.debug_symbol());
+                
+                self.compile_expr(symbol, unpack)?;
+                self.emit_instr(symbol, OpCode::IterInit);
+                self.emit_instr(symbol, OpCode::IterUnpack);
+                
+                if let Some(local_index) = unpack_len {
+                    self.emit_load_local_index(symbol, local_index);
+                    self.emit_instr(symbol, OpCode::Add);
+                    
+                    debug_assert!(self.scopes().is_temporary_scope());
+                    self.emit_end_scope();
+                }
+                
+                if static_len > 0 {
+                    self.compile_integer(symbol, static_len)?;
+                    self.emit_instr(symbol, OpCode::Add);
+                }
+                
+                Ok(Unpack::Dynamic)
+            }
+            
+            _ => {
+                self.compile_expr_with_symbol(last)?;
+                static_len = static_len.checked_add(1)
+                    .ok_or("unpack length limit exceeded")?;
+                    
+                if let Some(local_index) = unpack_len {
+                    self.emit_load_local_index(symbol, local_index);
+                    
+                    if static_len > 0 {
+                        self.compile_integer(symbol, static_len)?;
+                        self.emit_instr(symbol, OpCode::Add);
+                    }
+                    
+                    debug_assert!(self.scopes().is_temporary_scope());
+                    self.emit_end_scope();
+                    
+                    return Ok(Unpack::Dynamic)
+                }
+                
+                Ok(Unpack::Static(static_len))
+            }
+        }
+    }
+
     fn compile_atom(&mut self, symbol: Option<&DebugSymbol>, atom: &Atom) -> CompileResult<()> {
         match atom {
             Atom::Nil => self.emit_instr(symbol, OpCode::Nil),
@@ -874,9 +1002,23 @@ impl CodeGenerator<'_> {
             Atom::Group { modifier, inner } => {
                 // modifiers are not allowed outside of assignment
                 if let Some(modifier) = modifier {
-                    return Err("assignment modifier is not allowed here".into())
+                    return Err("assignment modifiers are not allowed outside of an assignment expression".into())
                 }
-                self.compile_expr(symbol, inner)?
+                
+                match &**inner {
+                    // tuple constructor
+                    Expr::Unpack(None) => return Err("need a value to unpack".into()),
+                    Expr::Unpack(Some(iter)) => {
+                        self.compile_expr(symbol, iter)?;
+                        self.emit_instr(symbol, OpCode::IterInit);
+                        self.emit_instr(symbol, OpCode::IterUnpack);
+                        self.emit_instr(symbol, OpCode::TupleN);
+                    }
+                    
+                    // parenthesized group
+                    expr => self.compile_expr(symbol, inner)?,
+                }
+                
             },
         }
         Ok(())
@@ -932,57 +1074,19 @@ impl CodeGenerator<'_> {
         Ok(())
     }
     
-    fn compile_invocation(&mut self, symbol: Option<&DebugSymbol>, mut args: &[ExprMeta]) -> CompileResult<()> {
+    fn compile_invocation(&mut self, symbol: Option<&DebugSymbol>, args: &[ExprMeta]) -> CompileResult<()> {
         // prepare argument list:
-        // [ callobj arg[n] arg[0] ... arg[n-1] nargs ] => [ ret_value ] 
+        // [ callobj arg[0] ... arg[n] nargs ] => [ ret_value ] 
 
-        let arg_len;
-        let mut unpack = None;
-        
         // process argument unpacking
-        if let Some((last, rest)) = args.split_last() {
-            let (expr, symbol) = last.clone().take();
-            match expr {
-                Expr::Ellipsis(None) => return Err("need a value to unpack".into()),
-                
-                Expr::Ellipsis(Some(inner)) => {
-                    args = rest;
-                    unpack.replace(ExprMeta::new(*inner, symbol));
-                }
-                
-                _ => { }
-            }
-            
-            if args.iter().any(|arg| matches!(arg.variant(), Expr::Ellipsis(..))) {
-                return Err("\"...\" can only be used to unpack the last argument".into());
-            }
-        }
-        
-        // compile arguments
-        if let Some((arg_last, args_rest)) = args.split_last() {
-            
-            self.compile_expr_with_symbol(arg_last)?;
-            for arg_expr in args_rest.iter() {
-                self.compile_expr_with_symbol(arg_expr)?;
-            }
-            
-            arg_len = u8::try_from(args.len())
-                .map_err(|_| "argument count limit exceeded")?;
-            
-        } else {
-            
-            arg_len = 0;
-        }
-        
-        self.emit_instr_byte(symbol, OpCode::UInt8, arg_len);
-        
-        if let Some(seq_expr) = unpack {
-            self.compile_expr_with_symbol(&seq_expr)?;
-            self.emit_instr(symbol, OpCode::CallUnpack);
-        } else {
-            self.emit_instr(symbol, OpCode::Call);
+        match self.compile_unpack_sequence(symbol, args)? {
+            Unpack::Empty => self.emit_instr_byte(symbol, OpCode::UInt8, 0),
+            Unpack::Static(len) => self.compile_integer(symbol, len)?,
+            Unpack::Dynamic => { } // nothing to do
         }
 
+        self.emit_instr(symbol, OpCode::Call);
+
         Ok(())
     }
     
@@ -1068,7 +1172,7 @@ impl CodeGenerator<'_> {
             
             LValue::Index(target) => unimplemented!(),
             
-            LValue::Tuple {..} | LValue::PackItem(..)
+            LValue::Tuple {..} | LValue::Pack(..)
                 => Err("can't update-assign to this".into()),
             
             LValue::Modifier {..} => unreachable!(),
@@ -1085,7 +1189,7 @@ impl CodeGenerator<'_> {
         match lhs {
             LValue::Tuple(items) => self.compile_assign_tuple(symbol, assign, items),
             
-            LValue::PackItem(..) => {
+            LValue::Pack(..) => {
                 let item = std::slice::from_ref(lhs);
                 self.compile_assign_tuple(symbol, assign, item)
             }
@@ -1129,7 +1233,7 @@ impl CodeGenerator<'_> {
     fn compile_decl_local_name(&mut self, symbol: Option<&DebugSymbol>, access: Access, name: InternSymbol) -> CompileResult<()> {
         
         match self.scopes_mut().insert_local(access, LocalName::Symbol(name))? {
-            InsertLocal::CreateNew => 
+            InsertLocal::CreateNew(..) => 
                 self.emit_instr(symbol, OpCode::InsertLocal),
             
             InsertLocal::HideExisting(local_index) =>
@@ -1139,13 +1243,6 @@ impl CodeGenerator<'_> {
         Ok(())
     }
     
-    // fn compile_decl_tuple(&mut self, symbol: Option<&DebugSymbol>, access: Access, targets: &[LValue]) -> CompileResult<()> {
-    //     self.compile_tuple_unpack(
-    //         symbol, targets, 
-    //         |self_, target| self_.compile_declaration(symbol, access, target)
-    //     )
-    // }
-    
     fn compile_assign_variable(&mut self, symbol: Option<&DebugSymbol>, lhs: &LValue, allow_nonlocal: bool) -> CompileResult<()> {
         
         match lhs {
@@ -1181,12 +1278,12 @@ impl CodeGenerator<'_> {
             }
             
             // nonlocal keyword is not required in the global frame
-            if !allow_nonlocal && !self.scopes().is_global_frame() {
+            if !allow_nonlocal && self.scopes().is_call_frame() {
                 return Err("can't assign to a non-local variable without the \"nonlocal\" keyword".into());
             }
             
             // check if an upvalue is found or can be created...
-            if !self.scopes().is_global_frame() {
+            if self.scopes().is_call_frame() {
                 if let Some(upval) = self.scopes_mut().resolve_or_create_upval(&local_name)? {
                     if !upval.mode().can_write() {
                         return Err("can't assign to immutable local variable".into());
@@ -1218,26 +1315,42 @@ impl CodeGenerator<'_> {
         }
     }
     
-    fn compile_assign_tuple(&mut self, symbol: Option<&DebugSymbol>, assign: AssignType, mut item_targets: &[LValue]) -> CompileResult<()> {
+    fn compile_assign_tuple(&mut self, symbol: Option<&DebugSymbol>, assign: AssignType, item_targets: &[LValue]) -> CompileResult<()> {
+        // process tuple packing patterns
         
-        let mut error_jump_sites = Vec::new();
+        let mut pack_targets = item_targets.iter().enumerate()
+            .filter_map(|(idx, target)| match target {
+                LValue::Pack(pack_target) => Some((idx, pack_target.as_deref())),
+                _ => None,
+            });
+            
+        let (idx, pack_target) = match pack_targets.next() {
+            Some(pack_target) => pack_target,
+            None => return self.compile_assign_tuple_nopack(symbol, assign, item_targets),
+        };
         
-        self.emit_instr(symbol, OpCode::IterInit);  // iterate to yield values
-        
-        // process tuple packing assingment
-        let mut pack_target = None;
-        
-        if let Some((LValue::PackItem(pack_item), rest)) = item_targets.split_last() {
-            pack_target.replace(pack_item);
-            item_targets = rest;
+        if !pack_targets.next().is_none() {
+            return Err("tuple assignment may contain only one \"...\" pack pattern".into())
         }
         
-        // compile to unrolled iteration
-        for (idx, target) in item_targets.iter().enumerate() {
-            
-            if matches!(target, LValue::PackItem(..)) {
-                return Err("\"...\" can only be used with the last item in tuple assignment".into());
-            }
+        let (pre_pack, rest) = item_targets.split_at(idx);
+        let (_, post_pack) = rest.split_at(1);
+        
+        self.compile_assign_tuple_pack(symbol, assign, pack_target, pre_pack, post_pack)
+    }
+    
+    fn compile_assign_tuple_pack(&mut self, symbol: Option<&DebugSymbol>, assign: AssignType, pack: Option<&LValue>, pre_pack: &[LValue], post_pack: &[LValue]) -> CompileResult<()> {
+        let mut error_jump_sites = Vec::new();
+        
+        // assignment needs to preserve original value for expression result
+        self.emit_instr(symbol, OpCode::Clone);
+        
+        // iterate to yield values
+        self.emit_instr(symbol, OpCode::IterInit);
+        
+        // compile to unrolled iteration for pre-pack items
+        for target in pre_pack.iter() {
+            debug_assert!(!matches!(target, LValue::Pack(..)));
             
             // check if there is an item left for this target
             let error_jump = self.emit_dummy_jump(symbol, Jump::IfFalse);
@@ -1250,53 +1363,135 @@ impl CodeGenerator<'_> {
             self.emit_instr(symbol, OpCode::Pop);
         }
         
-        let done_jump_site;
-        let pop_after;
-        if let Some(Some(pack_target)) = pack_target {
-            // exhaust the rest of the iterator and assign to pack_target
-            self.emit_instr(symbol, OpCode::IterUnpack);
-            self.emit_instr(symbol, OpCode::TupleN);
+        let temp_scope;
+        match (pack, post_pack) {
+            (None, []) => {
+                // if there are no post-pack items and no pack target just discard the iterator
+                self.emit_instr_byte(symbol, OpCode::Drop, 2);
+                temp_scope = false;
+            }
             
-            self.compile_assignment(symbol, assign, pack_target)?;
-            self.emit_instr(symbol, OpCode::Pop);
+            (Some(pack_target), []) => {
+                // exhaust the rest of the iterator and assign to pack_target
+                self.emit_instr(symbol, OpCode::IterUnpack);
+                self.emit_instr(symbol, OpCode::TupleN);
+                
+                self.compile_assignment(symbol, assign, pack_target)?;
+                self.emit_instr(symbol, OpCode::Pop);
+                temp_scope = false;
+            }
             
-            done_jump_site = self.emit_dummy_jump(symbol, Jump::Uncond);
-            pop_after = false;
-        
-        } else if let Some(None) = pack_target {
-            // just discard the iterator
-            self.emit_instr_byte(symbol, OpCode::Drop, 2);
-            done_jump_site = self.emit_dummy_jump(symbol, Jump::Uncond);
-            pop_after = false;
-            
-        } else {
-            // if the iterator is finished we've succeeded
-            done_jump_site = self.emit_dummy_jump(symbol, Jump::PopIfFalse);
-            pop_after = true; // pop the iterator
-            
-            // too many items
-            let message = format!("too many values to unpack (expected {})", item_targets.len());
-            self.emit_load_error(symbol, ErrorKind::UnpackError, message.as_str())?;
-            self.emit_instr(symbol, OpCode::Error);
+            (pack_target, post_pack) => {
+                // exhaust the rest of the iterator and assign the last items to post_pack
+                // then assign whatever is left to the pack_target
+                self.emit_instr(symbol, OpCode::IterUnpack);
+                
+                // calculate the pack length and store it
+                let post_len = IntType::try_from(post_pack.len())
+                    .map_err(|_| "unpack length limit exceeded")?;
+                
+                self.compile_integer(symbol, post_len)?;
+                self.emit_instr(symbol, OpCode::Sub);
+                
+                temp_scope = true;
+                self.emit_begin_scope(symbol, None, ScopeTag::Temporary);
+                let pack_len = self.emit_create_temporary(symbol, Access::ReadOnly)?;
+                
+                // check if there are enough items
+                self.emit_instr_byte(symbol, OpCode::UInt8, 0);
+                self.emit_instr(symbol, OpCode::LT);
+                let error_jump = self.emit_dummy_jump(symbol, Jump::PopIfTrue);
+                error_jump_sites.push(error_jump);
+                
+                // assign post-pack items
+                for target in post_pack.iter().rev() {
+                    debug_assert!(!matches!(target, LValue::Pack(..)));
+                    self.compile_assignment(symbol, assign, target)?;
+                    self.emit_instr(symbol, OpCode::Pop);
+                }
+                
+                // load the stored pack len and assign to pack target or discard
+                self.emit_load_local_index(symbol, pack_len);
+                if let Some(pack_target) = pack_target {
+                    self.emit_instr(symbol, OpCode::TupleN);
+                    self.compile_assignment(symbol, assign, pack_target)?;
+                    self.emit_instr(symbol, OpCode::Pop);
+                } else {
+                    self.emit_instr(symbol, OpCode::DropN);
+                }
+            }
         }
         
+        let done_jump_site = self.emit_dummy_jump(symbol, Jump::Uncond);
+        
         // not enough items
         let error_target = self.current_offset();
         for jump_site in error_jump_sites.iter() {
             self.patch_jump_instr(jump_site, error_target)?;
         }
         
-        let message = format!("not enough values to unpack (expected {})", item_targets.len());
+        let message = format!("not enough values to unpack (expected at least {})", pre_pack.len() + post_pack.len());
         self.emit_load_error(symbol, ErrorKind::UnpackError, message.as_str())?;
         self.emit_instr(symbol, OpCode::Error);
         
         // cleanup
         self.patch_jump_instr(&done_jump_site, self.current_offset())?;
         
-        if pop_after {
+        if temp_scope {
+            debug_assert!(self.scopes().is_temporary_scope());
+            self.emit_end_scope();
+        }
+        
+        Ok(())
+    }
+    
+    fn compile_assign_tuple_nopack(&mut self, symbol: Option<&DebugSymbol>, assign: AssignType, items: &[LValue]) -> CompileResult<()> {
+        let mut error_jump_sites = Vec::new();
+        
+        // assignment needs to preserve original value for expression result
+        self.emit_instr(symbol, OpCode::Clone);
+        
+        // iterate to yield values
+        self.emit_instr(symbol, OpCode::IterInit);
+        
+        // compile to unrolled iteration
+        for target in items.iter() {
+            debug_assert!(!matches!(target, LValue::Pack(..)));
+            
+            // check if there is an item left for this target
+            let error_jump = self.emit_dummy_jump(symbol, Jump::IfFalse);
+            error_jump_sites.push(error_jump);
+            
+            // advance the iterator and put the item on the stack
+            self.emit_instr(symbol, OpCode::IterNext);
+            
+            self.compile_assignment(symbol, assign, target)?;
             self.emit_instr(symbol, OpCode::Pop);
         }
         
+        // if the iterator is finished we've succeeded
+        let done_jump_site = self.emit_dummy_jump(symbol, Jump::PopIfFalse);
+        
+        // too many items
+        let message = format!("too many values to unpack (expected {})", items.len());
+        self.emit_load_error(symbol, ErrorKind::UnpackError, message.as_str())?;
+        self.emit_instr(symbol, OpCode::Error);
+        
+        // not enough items
+        let error_target = self.current_offset();
+        for jump_site in error_jump_sites.iter() {
+            self.patch_jump_instr(jump_site, error_target)?;
+        }
+        
+        let message = format!("not enough values to unpack (expected {})", items.len());
+        self.emit_load_error(symbol, ErrorKind::UnpackError, message.as_str())?;
+        self.emit_instr(symbol, OpCode::Error);
+        
+        // cleanup
+        self.patch_jump_instr(&done_jump_site, self.current_offset())?;
+        
+        self.emit_instr(symbol, OpCode::Pop);  // pop the iterator
+        
         Ok(())
     }
 }
@@ -1306,7 +1501,7 @@ impl CodeGenerator<'_> {
 
     fn compile_block_expression(&mut self, symbol: Option<&DebugSymbol>, label: Option<&Label>, suite: &ExprBlock) -> CompileResult<()> {
         
-        self.emit_begin_scope(symbol, ScopeTag::Block, label);
+        self.emit_begin_scope(symbol, label, ScopeTag::Block);
         self.compile_expr_block(symbol, suite)?;
         let block_scope = self.emit_end_scope();
         
@@ -1339,7 +1534,7 @@ impl CodeGenerator<'_> {
             // inside the statement list
             let branch_jump_site = self.emit_dummy_jump(symbol, Jump::IfFalse);
             
-            self.emit_begin_scope(symbol, ScopeTag::Branch, None);
+            self.emit_begin_scope(symbol, None, ScopeTag::Branch);
             self.compile_expr_block(symbol, branch.suite())?;
             self.emit_end_scope();
             
@@ -1357,7 +1552,7 @@ impl CodeGenerator<'_> {
         // else clause
         if let Some(suite) = else_clause {
             
-            self.emit_begin_scope(symbol, ScopeTag::Branch, None);
+            self.emit_begin_scope(symbol, None, ScopeTag::Branch);
             self.compile_expr_block(symbol, suite)?;
             self.emit_end_scope();
             
@@ -1468,31 +1663,31 @@ impl CodeGenerator<'_> {
     }
     
     fn compile_function_preamble(&mut self, symbol: Option<&DebugSymbol>, fundef: &FunctionDef) -> CompileResult<()> {
+        // process default and variadic arguments
+        // this ensures that exactly `signature.param_count()` values are on the stack
         
-        // define locals
         let signature = &fundef.signature;
+        if signature.param_count() == 0 {
+            return Ok(())
+        }
         
         for param in signature.required.iter() {
             self.scopes_mut().insert_local(param.mode, LocalName::Symbol(param.name))?;
-            //self.emit_instr(None, OpCode::InsertLocal);
         }
         
-        // will define local variables for all default arguments
         if !signature.default.is_empty() {
             self.compile_default_args(signature)?;
-            
             for param in signature.default.iter() {
                 self.scopes_mut().insert_local(param.mode, LocalName::Symbol(param.name))?;
-                // self.emit_instr(symbol, OpCode::InsertLocal);
             }
         }
         
         if let Some(param) = &signature.variadic {
             self.compile_variadic_arg(signature)?;
-
             self.scopes_mut().insert_local(param.mode, LocalName::Symbol(param.name))?;
-            // self.emit_instr(symbol, OpCode::InsertLocal);
         }
+        
+        self.emit_instr(symbol, OpCode::InsertArgs);
 
         Ok(())
     }
