@@ -16,30 +16,22 @@ pub type UpvalueIndex = u16;
                            // width set here so that the longest mnemonic is 16 chars
 const OP_NOP:              u8 = 0x00;
 const OP_EXIT:             u8 = 0x01;  // _ => !
-const OP_ERROR:            u8 = 0x02;  // [ error ] => !
+const OP_ERROR:            u8 = 0x02;  // T[ error ] => !
 
-const OP_RETURN:           u8 = 0x08;  // [ ...call frame... ret_value ] => [ ret_value ]
+const OP_RETURN:           u8 = 0x08;  // T[ ...call frame... ret_value ] => [ ret_value ]
 
-// The odd argument order is because the VM will swap "nargs" with "arg[n]" after reading "nargs".
-// This is because the function preamble will fill in any missing default arguments such that the
-// full argument list is always the same, and can be reasoned about at compile-time from the static signature.
-// This wouldn't be possible if nargs was in some arbitrary location in the middle of the argument list,
-// so swapping it to the beginning is the most efficient way to handle this.
-
-// [ callee arg[n] arg[0] ... arg[n-1] nargs ] => [ ret_value ] 
+// [ callee arg[0] ... arg[n] nargs ] => [ ret_value ] 
 const OP_CALL:             u8 = 0x09;
-
-// This instruction is needed because when unpacking, the final nargs value needs to be updated dynamically as we iterate.
-// [ callee arg[n] arg[0] ... arg[n-1] nargs arg_unpack ] => [ ret_value ]  -- nargs does not include arg_unpack! 
-const OP_CALL_UNPACK:      u8 = 0x0A;
+const OP_IN_ARGS:          u8 = 0x0A;
 
 // 0x10-17        Immediate Values
 
 const OP_POP:              u8 = 0x10;  // [ _ ] => []
 const OP_DROP:             u8 = 0x11;  // (u8); [ value[0] ... value[N] ] => []
-const OP_CLONE:            u8 = 0x12;  // [ value ] => [ value value ]
-const OP_SWAP:             u8 = 0x13;  // (u8); [ value[A] ... value[B] ] => [ value[B] ... value[A] ]
-const OP_SHIFT:            u8 = 0x14;  // (u8); [ value[A] ... value[B] ] => [ ... value[B] value[A] ]
+const OP_DROPN:            u8 = 0x12;  // [ value[0] ... value[N] N ] => []
+const OP_CLONE:            u8 = 0x13;  // [ value ] => [ value value ]
+const OP_SWAP:             u8 = 0x14;  // (u8); [ value[A] ... value[B] ] => [ value[B] ... value[A] ]
+const OP_SHIFT:            u8 = 0x15;  // (u8); [ value[A] ... value[B] ] => [ ... value[B] value[A] ]
 
 const OP_TUPLE:            u8 = 0x18;  // (u8); [ item[0] ... item[N] ] => [ tuple ]
 const OP_TUPLEN:           u8 = 0x19;  // [ item[0] ... item[N] N ] => [ tuple ]
@@ -65,12 +57,12 @@ const OP_ST_GLOBAL:        u8 = 0x4A;  // [ value name ] => [ value ]
 const OP_LD_GLOBAL:        u8 = 0x4B;  // [ name ] => [ value ]
 const OP_DP_GLOBAL:        u8 = 0x4C;  // [ name ] => []
 
-const OP_IN_LOCAL:         u8 = 0x50;  // [ value ] => [ value ]; vm.locals += 1
+const OP_IN_LOCAL:         u8 = 0x50;  // [ value ] => [ value ];
 const OP_ST_LOCAL:         u8 = 0x51;  // (u8);  [ value ] => [ value ]
 const OP_ST_LOCAL_16:      u8 = 0x52;  // (u16); [ value ] => [ value ]
 const OP_LD_LOCAL:         u8 = 0x53;  // (u8);  _ => [ value ]
 const OP_LD_LOCAL_16:      u8 = 0x54;  // (u16); _ => [ value ]
-const OP_DP_LOCALS:        u8 = 0x55;  // (u8); [ local[0] ... local[N] temporaries... ] => [ temporaries... ]; vm.locals -= N
+const OP_DP_LOCALS:        u8 = 0x55;  // (u8); vm.locals -= N
 
 const OP_ST_UPVAL:         u8 = 0x58;  // (u8);  [ value ] => [ value ]
 const OP_ST_UPVAL_16:      u8 = 0x59;  // (u16); [ value ] => [ value ]
@@ -152,10 +144,11 @@ pub enum OpCode {
     
     Return = OP_RETURN, 
     Call = OP_CALL,
-    CallUnpack = OP_CALL_UNPACK,
+    InsertArgs = OP_IN_ARGS,
     
     Pop = OP_POP,
     Drop = OP_DROP,
+    DropN = OP_DROPN,
     Clone = OP_CLONE,
     
     Tuple = OP_TUPLE,
@@ -248,10 +241,11 @@ impl OpCode {
             
             OP_RETURN => Self::Return,
             OP_CALL => Self::Call,
-            OP_CALL_UNPACK => Self::CallUnpack,
+            OP_IN_ARGS => Self::InsertArgs,
             
             OP_POP => Self::Pop,
             OP_DROP => Self::Drop,
+            OP_DROPN => Self::DropN,
             OP_CLONE => Self::Clone,
             
             OP_TUPLE => Self::Tuple,
@@ -409,10 +403,11 @@ impl core::fmt::Display for OpCode {
             
             Self::Return => "RETURN",
             Self::Call => "CALL",
-            Self::CallUnpack => "CALL_UNPACK",
+            Self::InsertArgs => "IN_ARGS",
             
             Self::Pop => "POP",
             Self::Drop => "DROP",
+            Self::DropN => "DROPN",
             Self::Clone => "CLONE",
             
             Self::Tuple => "TUPLE",
