@@ -15,10 +15,12 @@ pub(super) enum LocalName {
     Symbol(InternSymbol),
     
     // special local variables
+    
+    // these are created by the VM
     Receiver,  // inside a function call, this refers to the object that was called
     NArgs,     // inside a function call, the number of arguments passed at the call site
     
-    Temp,      // anonymous temporaries
+    Anonymous, // for internal temporaries. excluded from local variable resolution, they can only be referred to by local index
 }
 
 
@@ -39,16 +41,28 @@ impl Local {
 
 #[derive(Clone, Copy)]
 pub(super) enum InsertLocal {
-    CreateNew,
+    CreateNew(LocalIndex),
     HideExisting(LocalIndex),
 }
 
+impl From<InsertLocal> for LocalIndex {
+    fn from(result: InsertLocal) -> Self {
+        match result {
+            InsertLocal::CreateNew(local_index) => local_index,
+            InsertLocal::HideExisting(local_index) => local_index,
+        }
+    }
+}
+
+
 #[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
 pub(super) enum ScopeTag {
     Block,
     Loop,
     Branch,
     Function,
+    Global,
+    Temporary,
 }
 
 impl ScopeTag {
@@ -66,6 +80,13 @@ impl ScopeTag {
         }
     }
     
+    fn hide_from_nro(&self) -> bool {
+        match self {
+            Self::Temporary => true,
+            _ => false,
+        }
+    }
+    
     pub(super) fn is_expr_block(&self) -> bool {
         match self {
             Self::Block | Self:: Branch => true,
@@ -159,12 +180,15 @@ impl Scope {
     }
     
     fn find_local(&self, name: &LocalName) -> Option<&Local> {
+        if matches!(name, LocalName::Anonymous) {
+            return None; // anonymous temporaries should not be referenced
+        }
         self.locals.iter().find(|local| local.name == *name)
     }
     
     fn find_local_mut(&mut self, name: &LocalName) -> Option<&mut Local> {
-        if matches!(name, LocalName::Temp) {
-            return None; // temporaries should not be referenced
+        if matches!(name, LocalName::Anonymous) {
+            return None; // anonymous temporaries should not be referenced
         }
         self.locals.iter_mut().find(|local| local.name == *name)
     }
@@ -186,13 +210,18 @@ impl Scope {
     }
     
     fn insert_local(&mut self, mode: Access, name: LocalName) -> CompileResult<InsertLocal> {
+        // ensure only anonymous variables get inserted into hidden scopes
+        if self.tag.hide_from_nro() {
+            debug_assert!(name == LocalName::Anonymous);
+        }
+        
         // see if this local already exists in the current scope
         if let Some(mut local) = self.find_local_mut(&name) {
             (*local).mode = mode; // redeclare with new mutability
             Ok(InsertLocal::HideExisting(local.index))
         } else {
-            self.push_local(mode, name)?;
-            Ok(InsertLocal::CreateNew)
+            let local = self.push_local(mode, name)?;
+            Ok(InsertLocal::CreateNew(local.index))
         }
     }
 }
@@ -200,52 +229,69 @@ impl Scope {
 
 #[derive(Debug)]
 struct NestedScopes {
-    scopes: Vec<Scope>,
+    toplevel: Scope,
+    nested: Vec<Scope>,
 }
 
 impl NestedScopes {
-    fn new() -> Self {
-        Self { scopes: Vec::new() }
-    }
-    
-    fn is_empty(&self) -> bool {
-        self.scopes.is_empty()
-    }
-    
-    fn current_scope(&self) -> Option<&Scope> {
-        self.scopes.last()
-    }
-    
-    fn current_scope_mut(&mut self) -> Option<&mut Scope> {
-        self.scopes.last_mut()
-    }
-    
-    fn push_scope(&mut self, symbol: Option<&DebugSymbol>, tag: ScopeTag, label: Option<Label>) {
-        let prev_index = self.scopes.last().and_then(|scope| scope.last_index());
-        
-        let scope = Scope {
+    fn new(symbol: Option<&DebugSymbol>, tag: ScopeTag, label: Option<Label>) -> Self {
+        let toplevel = Scope {
             tag,
-            depth: self.scopes.len(),
-            prev_index,
+            depth: 0,
+            prev_index: None,
             symbol: symbol.copied(),
             locals: Vec::new(),
             control_flow: ControlFlowTracker::new(label),
         };
         
-        self.scopes.push(scope);
+        Self {
+            toplevel,
+            nested: Vec::new(),
+        }
+    }
+    
+    fn is_nested(&self) -> bool {
+        !self.nested.is_empty()
+    }
+    
+    fn current_scope(&self) -> &Scope {
+        self.nested.last().unwrap_or(&self.toplevel)
+    }
+    
+    fn current_scope_mut(&mut self) -> &mut Scope {
+        self.nested.last_mut().unwrap_or(&mut self.toplevel)
+    }
+    
+    fn push_scope(&mut self, symbol: Option<&DebugSymbol>, label: Option<Label>, tag: ScopeTag) {
+        let current_scope = self.current_scope();
+        
+        let scope = Scope {
+            tag,
+            depth: current_scope.depth() + 1,
+            prev_index: current_scope.last_index(),
+            symbol: symbol.copied(),
+            locals: Vec::new(),
+            control_flow: ControlFlowTracker::new(label),
+        };
+        
+        self.nested.push(scope);
     }
     
     fn pop_scope(&mut self) -> Scope {
-        self.scopes.pop().expect("pop empty scope")
+        self.nested.pop().expect("pop toplevel scope")
     }
     
     /// Iterate in name resolution order
     fn iter_nro(&self) -> impl Iterator<Item=&Scope> {
-        self.scopes.iter().rev()
+        self.nested.iter().rev()
+            .chain(std::iter::once(&self.toplevel))
+            .filter(|scope| !scope.tag().hide_from_nro())
     }
     
     fn iter_nro_mut(&mut self) -> impl Iterator<Item=&mut Scope> {
-        self.scopes.iter_mut().rev()
+        self.nested.iter_mut().rev()
+            .chain(std::iter::once(&mut self.toplevel))
+            .filter(|scope| !scope.tag().hide_from_nro())
     }
 }
 
@@ -274,11 +320,8 @@ pub(super) struct CallFrame {
 
 impl CallFrame {
     fn new(symbol: Option<&DebugSymbol>) -> Self {
-        let mut scopes = NestedScopes::new();
-        scopes.push_scope(symbol, ScopeTag::Function, None);
-        
         Self {
-            scopes,
+            scopes: NestedScopes::new(symbol, ScopeTag::Function, None),
             upvalues: Vec::new(),
         }
     }
@@ -341,17 +384,22 @@ pub(super) struct ScopeTracker {
 impl ScopeTracker {
     pub(super) fn new() -> Self {
         Self {
-            toplevel: NestedScopes::new(),
+            toplevel: NestedScopes::new(None, ScopeTag::Global, None),
             frames: Vec::new(),
         }
     }
     
     pub(super) fn is_global_scope(&self) -> bool {
-        self.frames.is_empty() && self.toplevel.is_empty()
+        // return true if the first non-temporary is global
+        self.get_current_scope(true).tag() == ScopeTag::Global
     }
     
-    pub(super) fn is_global_frame(&self) -> bool {
-        self.frames.is_empty()
+    pub(super) fn is_temporary_scope(&self) -> bool {
+        self.local_scopes().current_scope().tag() == ScopeTag::Temporary
+    }
+    
+    pub(super) fn is_call_frame(&self) -> bool {
+        !self.frames.is_empty()
     }
     
     pub(super) fn push_frame(&mut self, symbol: Option<&DebugSymbol>) {
@@ -372,31 +420,44 @@ impl ScopeTracker {
             .map_or(&mut self.toplevel, |frame| frame.scopes_mut())
     }
     
+    fn get_current_scope(&self, ignore_temp: bool) -> &Scope {
+        if ignore_temp {
+            self.local_scopes()
+                .iter_nro().next()
+                .expect("empty nro")
+        } else {
+            self.local_scopes().current_scope()
+        }
+    }
+    
+    fn get_current_scope_mut(&mut self, ignore_temp: bool) -> &mut Scope {
+        if ignore_temp {
+            self.local_scopes_mut()
+                .iter_nro_mut().next()
+                .expect("empty nro")
+        } else {
+            self.local_scopes_mut().current_scope_mut()
+        }
+    }
+    
     // scopes
     
-    pub(super) fn local_scope(&self) -> Option<&Scope> {
-        self.local_scopes().current_scope()
-    }
-    
-    pub(super) fn local_scope_mut(&mut self) -> Option<&mut Scope> {
-        self.local_scopes_mut().current_scope_mut()
-    }
-    
-    pub(super) fn push_scope(&mut self, symbol: Option<&DebugSymbol>, tag: ScopeTag, label: Option<Label>) {
-        self.local_scopes_mut().push_scope(symbol, tag, label);
+    pub(super) fn push_scope(&mut self, symbol: Option<&DebugSymbol>, label: Option<Label>, tag: ScopeTag) -> &mut Scope {
+        let local_scope = self.local_scopes_mut();
+        local_scope.push_scope(symbol, label, tag);
+        local_scope.current_scope_mut()
     }
     
     pub(super) fn pop_scope(&mut self) -> Scope {
         let scope = self.local_scopes_mut().pop_scope();
-        debug_assert!(self.frames.last().map_or(true, |frame| !frame.scopes().is_empty()), "pop last scope from call frame");
         scope
     }
     
     // local variables
     
     pub(super) fn insert_local(&mut self, mode: Access, name: LocalName) -> CompileResult<InsertLocal> {
-        let scope = self.local_scopes_mut().current_scope_mut().expect("insert local in global scope");
-        scope.insert_local(mode, name)
+        self.get_current_scope_mut(name != LocalName::Anonymous)
+            .insert_local(mode, name)
     }
     
     pub(super) fn resolve_local(&self, name: &LocalName) -> Option<&Local> {
