@@ -267,10 +267,8 @@ impl<I> Parser<'_, I> where I: Iterator<Item=Result<TokenMeta, LexerError>> {
             
             // expression statements
             _ => match self.parse_expr_variant(ctx)? {
-                
-                Expr::Ellipsis(..) => 
+                Expr::Unpack(None) =>
                     return Err("\"...\" is not allowed here".into()),
-                
                 expr => Stmt::Expression(expr)
             }
         };
@@ -577,7 +575,7 @@ impl<I> Parser<'_, I> where I: Iterator<Item=Result<TokenMeta, LexerError>> {
             
             let assign = Assignment {
                 lhs, op, rhs,
-                assign: assign.unwrap_or(AssignType::AssignLocal),
+                modifier: assign.unwrap_or(AssignType::AssignLocal),
             };
             
             Ok(Expr::Assignment(Box::new(assign)))
@@ -797,11 +795,6 @@ impl<I> Parser<'_, I> where I: Iterator<Item=Result<TokenMeta, LexerError>> {
             
             // Token::OpenBrace => Ok(ExprMeta::ObjectCtor(self.parse_object_constructor(ctx)?)),
             
-            Token::Echo => {
-                ctx.set_start(&self.advance().unwrap());
-                Expr::Echo(Box::new(self.parse_expr_variant(ctx)?))
-            },
-            
             _ => self.parse_unpack_expr(ctx)?,
         };
         Ok(expr)
@@ -812,7 +805,7 @@ impl<I> Parser<'_, I> where I: Iterator<Item=Result<TokenMeta, LexerError>> {
         let next = self.peek()?;
         if matches!(next.token, Token::Ellipsis) {
             ctx.set_end(&self.advance().unwrap());
-            return Ok(Expr::Ellipsis(None));
+            return Ok(Expr::Unpack(None));
         }
         
         let expr = self.parse_primary(ctx)?;
@@ -823,10 +816,10 @@ impl<I> Parser<'_, I> where I: Iterator<Item=Result<TokenMeta, LexerError>> {
             
             // Having multiple "..."s next to each other is really bad for readability
             // So require that they are separated by parens, e.g. (((foo...)...)...)
-            if matches!(expr, Expr::Ellipsis(..)) {
+            if matches!(expr, Expr::Unpack(..)) {
                 return Err("nested use of \"...\" must be enclosed in parentheses".into());
             }
-            return Ok(Expr::Ellipsis(Some(Box::new(expr))));
+            return Ok(Expr::Unpack(Some(Box::new(expr))));
         }
         
         Ok(expr)
@@ -943,7 +936,7 @@ impl<I> Parser<'_, I> where I: Iterator<Item=Result<TokenMeta, LexerError>> {
             }
             
             let fun_decl = Assignment {
-                assign: AssignType::DeclImmutable,
+                modifier: AssignType::DeclImmutable,
                 op: None,
                 lhs: lvalue,
                 rhs: Expr::FunctionDef(function_def),
@@ -1380,10 +1373,15 @@ impl<I> Parser<'_, I> where I: Iterator<Item=Result<TokenMeta, LexerError>> {
         }
 
         // Check for lvalue modifier
-        let modifier = self.try_parse_assign_keyword(ctx)?;
+        let mut modifier = self.try_parse_assign_keyword(ctx)?;
 
         // Parse inner expression
-        let expr = self.parse_expr_variant(ctx)?;
+        let mut expr = self.parse_expr_variant(ctx)?;
+        
+        // if inner expression is an assignment, transfer our modifier to it
+        if let Expr::Assignment(assign) = &mut expr {
+            assign.modifier = modifier.take().unwrap_or(assign.modifier);
+        }
         
         // Consume and check closing paren
         let next = self.advance()?;
