@@ -13,13 +13,14 @@ pub enum AssignType {
     DeclMutable,
 }
 
+// TODO rename to Pattern?
 #[derive(Debug, Clone)]
 pub enum LValue {
     Identifier(InternSymbol),
     Attribute(Box<AttributeTarget>), // receiver, attribute name
     Index(Box<IndexTarget>), // receiver, index expression
     Tuple(Box<[LValue]>),
-    PackItem(Option<Box<LValue>>),
+    Pack(Option<Box<LValue>>),
     
     Modifier {
         modifier: AssignType,
@@ -46,7 +47,7 @@ pub struct IndexTarget {
 #[derive(Debug, Clone)]
 pub struct Assignment {
     pub lhs: LValue,
-    pub assign: AssignType,
+    pub modifier: AssignType,
     pub op: Option<BinaryOp>, // e.g. for +=, -=, *=, ...
     pub rhs: Expr,
 }
@@ -114,11 +115,11 @@ impl TryFrom<Expr> for LValue {
             
             Expr::Primary(primary) => primary.try_into(),
             
-            Expr::Ellipsis(Some(expr)) => {
+            Expr::Unpack(Some(expr)) => {
                 let inner = LValue::try_from(*expr)?;
-                Ok(Self::PackItem(Some(Box::new(inner))))
+                Ok(Self::Pack(Some(Box::new(inner))))
             }
-            Expr::Ellipsis(None) => Ok(Self::PackItem(None)),
+            Expr::Unpack(None) => Ok(Self::Pack(None)),
             
             Expr::Tuple(items) if !items.is_empty() => {
                 let mut lvalue_items = Vec::new();
